from superfish.parsers import parse_fish_t7, parse_poisson_t7

from pmd_beamphysics import FieldMesh

import numpy as np

from glob import glob
import os


def get_t7(path):
    """
    Returns a list of T7 files in path
    """
    return glob(os.path.join(path, '*T7'))


def interpolate2d(sf,
                  zmin=-1000, zmax=1000, nz=100,
                  rmin=0, rmax=0, nr=1, return_fieldmesh=False):
    """
    
    Runs SF7.EXE on a Superfish object sf, requesting interpolating data, and reads the Parmela T7 file. 
    
    Units are in the input units of the program. 
    
    Labels will label the output columns. The user must know what these are from the problem.
    
    TODO: detect these from SF object
    
    SF7 automatically adjusts the bounds if they are requested outside of the computational domain.
    
    Returns a dict with:
        rmin: minimum radius in cm
        rmax: maximum radius in cm
        nr:   number of radius points
        zmin: minimum z in cm
        zmax: maximum z in cm
        nz:   number of z points
        freq: frequency in MHz
        data: 2D array of shape (nr, nz)
    
    
    """
    
    problem = sf.problem 
    
    # fish and poisson have the opposite conventions:
    if problem == 'fish':
    # The interpolation grid
        F=f"""Parmela
{zmin} {rmin} {zmax} {rmax}
{nz-1} {nr-1}
End"""
        
    elif problem == 'poisson':
        F=f"""Parmela
{rmin} {zmin} {rmax} {zmax}
{nr-1} {nz-1}
End"""
    else:
        raise ValueError(f'unknowm problem: {problem}')
    

    # Clear old T7
    for f in get_t7(sf.path):
        os.remove(f)
    
    # Write 
    ifile = sf.basename+'.IN7'
    with open(os.path.join(sf.path, ifile), 'w') as f:
        f.write(F)
    
    # Needed so that the fields aren't normalized to 1 MV/m average
    inifile = os.path.join(sf.path, 'SF.INI')
    with open(inifile, 'w') as f:
        f.write("""[global]
Force1MVperMeter=No""")
    
    # Run
    sf.run_cmd('sf7', ifile)
    
    # Get the filename
    t7file = get_t7(sf.path)
    assert len(t7file) == 1, f'T7 file is missing.'
    t7file = t7file[0]
    
    
    # Optional fieldmesh parsing
    if return_fieldmesh:
        # Parsing is different for each:
        if problem == 'fish':
            type = 'electric'
            return FieldMesh.from_superfish(t7file)
        else:
            if sf.output['sfo']['header']['variable']['XJFACT'] == 0:
                type = 'electric'
            else:
                type = 'magnetic'
            return FieldMesh.from_superfish(t7file, type=type)
        
        
    
    # Parsing is different for each:
    if problem == 'fish':
        type = 'electric'
        d =  parse_fish_t7(t7file)
    else:
        if sf.output['sfo']['header']['variable']['XJFACT'] == 0:
            type = 'electric'
        else:
            type = 'magnetic'
        d =  parse_poisson_t7(t7file, type=type)
    
    return d


import numpy as np
import scipy.constants
mu_0 = scipy.constants.mu_0

import warnings

def fish_externalfield_data(t7data,
                      eleAnchorPt = 'beginning',
                      fieldScale = 1,
                      RFphase = 0,
                      z_offset = 0,
                      name = None,
                      normalize_by_Ez0=False
                      ):
    """
    Input: 
        t7data dict from a parsed T7 file. This is in the native Superfish units. 
    
    Superfish fields oscillate as:
        Er, Ez ~ cos(wt)
        Hphi   ~ -sin(wt)
     
    
    For complex fields oscillating as e^-iwt
    
        Re(Ex*e^-iwt)   ~ cos
        Re(-iB*e^-iwt) ~ -sin        
    and therefore B = -i * mu_0 * H_phi is the complex magnetic field in Tesla
    
    Output:
        dict with dicts:
            attrs
            components
            
    that are ready to be written to an HDF5 file. 
        
    
    """
    warnings.warn('fish_externalfield_data should be replaced by direct parsing using openPMD-beamphyics', DeprecationWarning)
    
    attrs = {}
    attrs['eleAnchorPt'] = eleAnchorPt
    
    # Use these to calculate spacing
    zmin = t7data['zmin']*.01
    zmax = t7data['zmax']*.01
    rmin = t7data['rmin']*.01
    rmax = t7data['rmax']*.01
    
    assert rmin == 0, f'rmin is not zero: {rmin}'
    
    nr = t7data['nr']
    nz = t7data['nz']
    
    dz = (zmax-zmin)/(nz-1)
    dr = (rmax)/(nr-1)
    
    attrs['gridGeometry'] = 'cylindrical'
    attrs['axisLabels'] = ('r', 'theta', 'z')
    attrs['gridLowerBound'] = (0, 1, 0)
    attrs['gridSize'] = (nr, 1, nz)        
    attrs['gridSpacing'] = (dr, 0, dz)
    
    # Set requested zmin
    attrs['gridOriginOffset'] = (0, 0, zmin+z_offset)
    
    attrs['fundamentalFrequency'] = t7data['freq']*1e6
    
    # Bmad non-standard
    ## attrs['masterParameter'] = 'VOLTAGE'
    
    attrs['harmonic'] = 1
    attrs['RFphase'] = RFphase
    
    if name:
        attrs['name'] = name     
            
    # Normalize on-axis field
    if normalize_by_Ez0:
        Ez0_max = 1e6*np.abs(t7data['Ez'][0,:]).max() # V/m
    else:
        Ez0_max = 1
        
    components = {}
    components['electricField/r'] = t7data['Er'].reshape(nr, 1, nz).astype(np.complex) * 1e6/Ez0_max
    components['electricField/z'] = t7data['Ez'].reshape(nr, 1, nz).astype(np.complex) * 1e6/Ez0_max
    components['magneticField/theta'] = t7data['Hphi'].reshape(nr, 1, nz) * -1j*mu_0/Ez0_max # -i * mu_0
    
    
    
    return dict(attrs=attrs, components=components)



def poisson_externalfield_data(t7data,
                      eleAnchorPt = 'beginning',
                      fieldScale = 1,
                      type='electric',
                      z_offset=0,
                      name = None,
                      normalize_by_fz0=False                               
                      ):
    """
    Input: 
        t7data dict from a parsed T7 file. This is in the native Superfish units. 
    
    Output:
        dict with dicts:
            attrs
            components
        
    that are ready to be written to an HDF5 file. 
        
    
    """
    
    attrs = {}
    
    assert eleAnchorPt in ['beginning', 'center', 'end'], f'Unallowed eleAnchorPt point: {eleAnchorPt}'
    
    attrs['eleAnchorPt'] = eleAnchorPt
        
    # Use these to calculate spacing
    zmin = t7data['zmin']*.01
    zmax = t7data['zmax']*.01
    rmin = t7data['rmin']*.01
    rmax = t7data['rmax']*.01
    
    assert rmin == 0, f'rmin is not zero: {rmin}'
    
    nr = t7data['nr']
    nz = t7data['nz']
    
    dz = (zmax-zmin)/(nz-1)
    dr = (rmax)/(nr-1)
    
    attrs['gridGeometry'] = 'cylindrical'
    attrs['axisLabels'] = ('r', 'theta', 'z')
    attrs['gridLowerBound'] = (0, 1, 0)
    attrs['gridSize'] = (nr, 1, nz)        
    attrs['gridSpacing'] = (dr, 0, dz)
    
    # Set requested zmin
    attrs['gridOriginOffset'] = (0, 0, zmin+z_offset)    
    
    attrs['fundamentalFrequency'] = 0
    attrs['harmonic'] = 0
    attrs['RFphase'] = 0
    
    if name:
        attrs['name'] = name     
            
    if type=='electric':
        fr = 'Er'
        fz = 'Ez'  
        ofr = 'electricField/r' # Ouptut names
        ofz = 'electricField/z'
        factor = 1 # V/m
         # Bmad non-standard
         # attrs['masterParameter'] = 'VOLTAGE'     
    elif type=='magnetic':
        fr = 'Br'
        fz = 'Bz' 
        ofr = 'magneticField/r'
        ofz = 'magneticField/z'
        factor = 1e-4 # G -> T
        # Bmad non-standard
        # attrs['masterParameter'] = 'BS_FIELD'
    else:
        raise ValueError(f"Unknown type: {type}. Allowed: 'electric' or 'magnetic'")
            
    # Normalize on-axis field     
    if normalize_by_fz0:
        fz0_max = np.abs(t7data[fz][0,:]*factor).max()            
    else:
        fz0_max = 1
        
    components = {}
    components[ofr] = t7data[fr].reshape(nr, 1, nz)*factor/fz0_max
    components[ofz] = t7data[fz].reshape(nr, 1, nz)*factor/fz0_max    

    return dict(attrs=attrs, components=components)
    

def write_fish_t7(filename, t7data,  fmt='%10.8e'):
    """
    Writes a T7 file from FISH t7data dict. 
    
    
    See:
        superfish.parsers.parse_fish_t7
    """
    
    # Collect these
    xmin = t7data['zmin']
    xmax = t7data['zmax']
    nx   = t7data['nz']
    ymin = t7data['rmin']
    ymax = t7data['rmax']
    ny   = t7data['nr']
    freq = t7data['freq']
    
    
    header = f"""{xmin} {xmax} {nx-1}
{freq}
{ymin} {ymax} {ny-1}"""
    
    
    # Unroll the arrays
    dat = np.array([t7data[f].reshape(nx*ny).T for f in ['Ez', 'Er', 'E', 'Hphi']]).T
    
    np.savetxt(filename, dat, header=header, comments='',  fmt = fmt)
    
    return filename

def write_poisson_t7(filename, t7data,  fmt='%10.8e'):
    """
    Writes a T7 file from POISSON t7data dict. 
    
    
    See:
        superfish.parsers.parse_poisson_t7
    """
    
    # Collect these
    ymin = t7data['zmin']
    ymax = t7data['zmax']
    ny   = t7data['nz']
    xmin = t7data['rmin']
    xmax = t7data['rmax']
    nx   = t7data['nr']
    #freq = t7data['freq']
    
    
    header = f"""{xmin} {xmax} {nx-1}
{ymin} {ymax} {ny-1}"""
    
    
    if 'Ez' in t7data:
        keys = ['Er', 'Ez']
    else:
        keys = ['Br', 'Bz']
    
    # Unroll the arrays
    dat = np.array([t7data[f].reshape(nx*ny, order='F').T for f in keys]).T
    
    np.savetxt(filename, dat, header=header, comments='',  fmt = fmt)
    
    return filename


from . import parsers
from .plot import plot_wall
from .interpolate import interpolate2d

import os
import platform
import subprocess
import tempfile
from time import time
import shutil


class Superfish:
    
    # Class attributes for the container
    _container_image = 'hhslepicka/poisson-superfish:latest'
    _windows_exe_path = 'C:\\LANL\\'  # Windows only'
    _singularity_image = '~/poisson-superfish_latest.sif'

    # Automatically detect the container method
    if shutil.which('docker'):
        _container_command = 'docker run {interactive_flags} --rm -v {local_path}:/data/ {image} {cmds}'
    elif shutil.which('shifter'):
        _container_command = 'shifter --image={image} {cmds}'
    elif shutil.which('singularity'):
        _container_command = 'singularity exec {singularity_image} {cmds}'
    else:
        _container_command = None

    def __init__(
            self,
            automesh=None,
            problem='fish',
            use_tempdir=True,
            use_container='auto',
            interactive=False,
            workdir=None,
            verbose=True):
        """
        Poisson-Superfish object
        
        
        """
        self.configured = False
        self.problem = problem
            
        self.verbose = verbose
        self.use_tempdir = use_tempdir
        self.interactive = interactive
        if workdir:
            workdir = os.path.abspath(workdir)
            assert os.path.exists(workdir), f'workdir does not exist: {workdir}'           
        self.workdir = workdir
            
        if automesh:
            self.load_input(automesh)
            self.configure()

        if use_container == 'auto':
            if platform.system() == 'Windows' and os.path.exists(self._windows_exe_path):
                self.use_container = False
                self.vprint(f'Using executables installed in {self._windows_exe_path}')
                
            else:
                self.vprint(f'Using container on {platform.system()}:')
                self.vprint('    ', self._container_command)
                self.use_container = True
            
        else:
            self.use_container = use_container

    @property
    def basename(self):
        return self.input['basename']

    @property
    def automesh_name(self):
        return self.basename+'.AM'
        
    def param(self, key):
        """
        Parameters (Variables and Constants) from the readback in the SFO file. 
        """
        return self.output['sfo']['header']['variable'][key]
  
    def param_info(self, key):
        """
        Returns the description of a paramater
        """
        return self.output['sfo']['header']['description'][key]
    
        
    def configure(self):     
        """
        Configures paths to run in.
        """
        
        # Set paths
        if self.use_tempdir:

            # Need to attach this to the object. Otherwise it will go out of scope.
            self.tempdir = tempfile.TemporaryDirectory(dir=self.workdir)
            self.path = self.tempdir.name
            
        else:
            
            if self.workdir:
                self.path = self.workdir
                self.tempdir = None
            else:
                # Work in place
                self.path = self.original_path        
     
        self.vprint('Configured to run in:', self.path)
        
        self.configured = True  

    def fieldmesh(self, 
                  zmin=-100, 
                  zmax=100, 
                  nz=0,
                  dz=0,
                  rmin=0, 
                  rmax=100, 
                  nr=0,
                  dr=0):
        """
        Interpolates field over a grid. Similar to .interpolate, 
        but input units are in meters. 
        
        Various combinations of the spacings and grid point numbers
        nz, dz, nr, nz can be used.
        If nether is specified, a default of 100 grid points will be used.
        
        Returns an openPMD-beamphysics FieldMesh object
        """
    
        conv = self.param('CONV')
        fac = 100/conv
        
        # Various input possibilities for the grid
        if dz and nz:
            zmax = zmin + (nz-1)*dz
        elif dz and not nz:
            nz = int((zmax-zmin)/dz)+1
            zmax = zmin + (nz-1)*dz
        elif not dz and not nz:
            # Default
            nz = 100
            
        if dr and nr:
            rmax = rmin + (nr-1)*dr
        elif dr and not nr:
            nr = int((rmax-rmin)/dr)+1
            rmax = rmin + (nr-1)*dr           
        elif not dr and not nr:
            # Default
            nr = 100

        FM = interpolate2d(
            self,
            zmin=zmin*fac, zmax=zmax*fac, nz=nz,
            rmin=rmin*fac, rmax=rmax*fac, nr=nr,
            return_fieldmesh=True
        )
        
        return FM        
        
    def interpolate(self, zmin=-1000, zmax=1000, nz=100, rmin=0, rmax=0, nr=1):
        """
        Interpolates field over a grid. 
        """
    
        t7data = interpolate2d(
            self,
            zmin=zmin, zmax=zmax, nz=nz,
            rmin=rmin, rmax=rmax, nr=nr
        )
        
        return t7data

    def run(self):
        """
        Writes input, runs autofish, and loads output
        
        """
        
        assert self.configured, 'not configured to run'
        
        self.write_input()
        
        t0 = time()
        
        if self.problem == 'fish':
            self.run_cmd('autofish', self.automesh_name)    
        else:
            self.run_cmd('automesh', self.automesh_name) 
            self.run_cmd('poisson')    
            self.run_cmd('sfo')    
           
        dt = time() - t0
        self.vprint(f'Done in {dt:10.2f} seconds')
        
        self.load_output()
        
    def container_run_cmd(self, *args):
        """
        Returns the run command string for the container.
        
        The container data should live in its /data/ folder.
        
        """
        
        cmds = ' '.join(args)

        if self.interactive:
            assert platform.system() == 'Darwin', 'TODO interactive non-Darwin'
            cmd0 = "IP=$(ifconfig en0 | grep inet | awk '$1==\"inet\" {print $2}');xhost + $IP;"
            interactive_flags = "-e INTERACTIVE_FISH=1 -e DISPLAY=$IP:0"
        else:
            cmd0 = ''
            interactive_flags = ''

        cmd = self._container_command.format(
            local_path=self.path,
            image=self._container_image,
            interactive_flags=interactive_flags,
            cmds=cmds,
            singularity_image=self._singularity_image
        )
        
        return cmd0+cmd
    
    def windows_run_cmd(self, *args):
        
        cmd = os.path.join(self._windows_exe_path, args[0].upper()+'.EXE')
        
        assert os.path.exists(cmd), f'EXE does not exist: {cmd}'
        
        if len(args) > 1:
            cmd = cmd + ' ' + ' '.join(args[1:])
       
        return cmd
        
    def run_cmd(self, *cmds, **kwargs):
        """
        Runs a command in in self.
        
        Example:
            .run_cmd(['C:\LANL\AUTOMESH.EXE', 'TEST.AM'], timeout=1)
        
        """
        if platform.system() == 'Windows':
            cmds = self.windows_run_cmd(*cmds)
        else:
            cmds = self.container_run_cmd(*cmds)
        
        self.vprint(f'Running: {cmds}')
    
        logfile = os.path.join(self.path, 'output.log')
    
        if self.use_container:
            
            # Shifter doesn't need volume mounting
            if (self._container_command.startswith('shifter') or
                    self._container_command.startswith('singularity')):
                cwd = self.path
            else:
                cwd = None
           
            with open(logfile, "a") as output:
                P = subprocess.call(
                    cmds, shell=True, stdout=output, stderr=output, cwd=cwd,
                    **kwargs
                )
        else:
            # Windows needs this
            P = subprocess.run(cmds.split(), cwd=self.path, **kwargs)
            
        return P
    
        ## Actual run
        #P = subprocess.run(cmds, shell=True, cwd=self.path, **kwargs)
        #
        #print(P)
        #
        #return P
    
    def load_input(self, input_filePath):
        """
        
        """
        f = os.path.abspath(input_filePath)
        
        # Get basename. Should be upper case to be consistent with output files (that are always upper case)
        _, fname = os.path.split(f) 
        basename = os.path.splitext(fname)[0].upper()
        self.input = dict(basename=basename)

        self.input['automesh'] = parsers.parse_automesh(f)  

    def load_output(self):
        """
        Loads SFO output file
        """
        
        self.output={}
        
        sfofile = os.path.join(self.path, self.basename+'.SFO')
        
        if not os.path.exists(sfofile):
            self.vprint('Warking: no SFO file to load.')
            return
        
        self.output['sfo'] = parsers.parse_sfo(sfofile)
        
        self.vprint('Parsed output:', sfofile)

    def plot_wall(self, units='original', **kwargs):
        if units == 'original':
            conv = 1
        elif units == 'cm':
            conv = self.param('CONV')
        else:
            raise ValueError(f'Units must be original or cm: {units}')
            
        plot_wall(self.output['sfo']['wall_segments'], conv=conv, **kwargs)

    def write_input(self):
        """
        Writes automesh input from .input['automesh']
        """
        
        file = os.path.join(self.path, self.input['basename']+'.AM')
        with open(file, 'w') as f:
            for line in self.input['automesh']:
                f.write(line)

    def vprint(self, *args):
        """verbose print"""
        if self.verbose:
            print(*args)   

    def __repr__(self):
        memloc = hex(id(self))
        if self.configured:
            return f'<Superfish configured to run in {self.path}>' 
        
        return f'<Superfish at {memloc}>'        



import matplotlib.pyplot as plt
import matplotlib
from copy import copy

# Get a default color map
CMAP0 = copy(plt.get_cmap('plasma'))
CMAP0.set_under('white')

from mpl_toolkits.axes_grid1 import make_axes_locatable

import numpy as np


def add_t7data_to_axes(t7data, ax, field='E', cmap=None, vmin=1e-19, scale=1):
    """
    Adds a field from t7data to an axes ax. 
    
    scale will scale the extents, to account for different units.
    Example: scale=10 will place this data in cm on a plot in mm
    
    """

    
    extent = [t7data[k]*scale for k in ('zmin', 'zmax', 'rmin', 'rmax')]
    
    if not cmap:
        cmap = CMAP0
        
    if field in ('E', 'B') and field not in t7data:
        data = np.hypot(t7data[field+'r'], t7data[field+'z'])
    else:
        data = t7data[field]
        
    
    ax.imshow(np.flipud(data), extent=extent, cmap=cmap, vmin=vmin )
    
    return ax




def perp(x, y, scale=1):
    """
    Calculates perpendicular lines.
    
    Returns two np.array: PX, PY 
        where each has the shape: (len(x)-1, 2)
    
    To test:
    
        X = np.array([1,2,3,4,8])
        Y = np.array([4,5,6,9,10])
        
        fig, ax = plt.subplots()
        ax.set_aspect('equal')
        ax.plot(X, Y)
        for i in range(len(PX)):
           ax.plot(PX[i,:], PY[i,:])
        
    """
    dx = np.diff(x)
    dy = np.diff(y)
    norm = np.hypot(dx, dy)
    
    # perp
    px0 = x[:-1]+dx/2
    py0 = y[:-1]+dy/2
    px1 = px0 +dy/norm *scale
    py1 = py0 -dx/norm *scale
    
    return np.array([px0, px1]).T, np.array([py0, py1]).T

def add_wall_segment_to_axes(seg, ax, perp_scale=0, max_field=1, field='E', cmap=None, conv=1):
    """
    Adds wall segments to axes. 
    
    If perp_scal >0, the field strength will be drawn as a perpendicular line. 
    max_field and the color map are needed to color these
    
    conv is the unit conversion factor used by Superfish. 
    
    """
    
    x = seg['wall']['Z']*conv
    y = seg['wall']['R']*conv

    # Wall segment
    ax.plot(x, y, color='black')
        
    if perp_scale:
        
        # Fetch the field
        F = seg['wall'][field]/max_field # field, E or B
    
        scales = perp_scale*F[:-1]
    
        px, py = perp(x, y, scale=scales)
        
        if not cmap:
            cmap = CMAP0    
        
        color = cmap(F[:-1])   

        # Draw perp lines
        for i in range(len(px)):
               ax.plot(px[i,:], py[i,:], color=color[i])
                
                
def plot_wall(wall_segments, 
              perp_scale=0, 
              field='E', 
              max_field=None,
              cmap=None,
              ax = None,
              conv=1,
              return_figure=False,
              
              **kwargs):
    """
    Plots the wall from wall segments.
    
    If perp_scale > 0, the field will be plotted
    
    conv is the conversion factor to internal units. If given, the 
    
    TODO: this is only for FISH problems so far
    
    """
    
    if not ax:
        fig, ax = plt.subplots( **kwargs)
    
    if not cmap:
        cmap = plt.get_cmap('plasma')    
    
    if perp_scale and not max_field:
        max_field = np.array([seg['wall'][field].max() for seg in wall_segments]).max()
    else:
        max_field = 0
    
    for seg in wall_segments:
        add_wall_segment_to_axes(seg, ax, perp_scale=perp_scale,
                                 field=field,
                                 max_field=max_field, cmap=cmap, 
                                 conv=conv)
      
    # Labels and units
    units = wall_segments[0]['units']

    ax.set_aspect('equal')
    if conv == 1:
        ax.set_xlabel('z '+units['Z'])
        ax.set_ylabel('r '+units['R'])  
        
    else:
        ax.set_xlabel('z (cm)')
        ax.set_ylabel('r (cm)')    
    
    
    if perp_scale == 0:
        if return_figure:
            return fig
        else:
            return
        
    divider = make_axes_locatable(ax)
    cax = divider.append_axes('right', size='5%', pad=0.05)
    
    # Legend units
    field_unit = units[field]
    
    # Tweak for convenience
    if field_unit=='(V/m)' and max_field >1e6:
        sc = 1e-6
        field_unit='(MV/m)'
    elif field_unit=='(V/cm)' and max_field >1e4:
        sc = 1e-4
        field_unit='(MV/m)'        
        
    else:
        sc = 1
     
    # Add legend
    norm = matplotlib.colors.Normalize(vmin=0, vmax=max_field*sc)
    fig.colorbar(matplotlib.cm.ScalarMappable(norm=norm, cmap=cmap),
             cax=cax, orientation='vertical', label=f'|{field}| max {field_unit}')             
    
    if return_figure:
        return fig
    
    
#plot_wall(SF.output['sfo']['wall_segments'], perp_scale=0, field='H', figsize=(20,8))                








import numpy as np
import os

def parse_automesh(file):
    if os.path.exists(file):
        lines = open(file, 'r').readlines()
    return lines


def parse_sfo(filename, verbose=False):
    """
    Master parser for the SFO file.

    Returns the output dict.

    """

    groups = parse_sfo_into_groups(filename)


    d = {
        'wall_segments':[],
        'other':{}
    }
    segments = []

    for g in groups:
        dat = process_group(g, verbose=verbose)
        type = dat['type']

        if type == 'wall_segment':
            d['wall_segments'].append(dat)
        elif type in ['summary','BeamEnergy']:
            d[type] = dat
        elif type == 'header':
            d['header'] = parse_header_lines(dat['lines'])
        
        else:
            d['other'][type] = dat
            
    # update Kinetic energy in 'summary' with the right value
    if 'summary' in d and 'BeamEnergy' in d:
        d['summary']['data']['kinetic_energy'] = d['BeamEnergy']['data']['BeamEnergy']/1e6
        d['summary']['units']['kinetic_energy'] = 'MeV'

    return d




def parse_sfo_into_groups(filename):
    """
    Parses SFO file into groups according to separator that starts with:
    '-------------------'

    Returns a list of dicts, with:
        raw_type: the first line
        lines: list of lines

    """

        
    with open(filename, 'r') as f:
        lines = f.readlines()
    
    groups = []

    sep = '-------------------'
    
    
    g = {'raw_type':'header', 'lines':[]}
    groups = [g]
    
    new_group = False
    
    for line in lines:
        line = line.strip()
        
        # Skip empty lines
        if not line:
            continue
            
        # Look for new group    
        if line.startswith(sep):
            new_group = True
            continue
        
        # Check for new group
        if new_group:
            gname = line
            new_group = False
            g = {'raw_type': gname, 'lines':[]}
            groups.append(g)
            continue
            
        # regular line
        g['lines'].append(line)

    return groups


def process_group(group, verbose=False):
    """
    processes a single output group dict into usable data.
    """

    rtype = group['raw_type']
    lines = group['lines']

    d = {}

    if rtype.startswith('All calculated values below refer to the mesh geometry only'):
        d['type'] = 'summary'
        d['data'], d['units'] = parse_sfo_summary_group(lines)
    elif rtype.startswith('Power and fields on wall segment') or rtype.startswith('Fields on segment'):
        d['type'] = 'wall_segment'
        line1 = rtype # This should be parsed fully

        d.update(parse_sfo_segment([line1]+lines))
    elif rtype.startswith('The field normalization factor ASCALE for this problem is based'):
        d['type'] = 'BeamEnergy'
        d['data'], d['units'] = parse_sfo_beam_energy(lines)

    else:
        # No parser yet:
        if verbose:
            print('No parser for:', rtype)

        d['type'] = rtype
        d['lines'] = lines

    return d



#_________________________________
# T7 files

def parse_fish_t7(t7file, geometry='cylindrical'):
    """
    Parses a T7 file. The T7 header should have:
    
    xmin(cm), xmax(cm), nx-1
    freq(MHz)
    ymin(cm), ymax(cm), ny-1
    4 columns of data: Ez, Er, E, Hphi
    
    TODO: Poisson problems, detect rectangular or cylindrical coordinates
    
    Returns a dict with:
        rmin
        rmax
        nr
        zmin
        zmax
        nz
        freq: frequency in MHz
        data: 2D array of shape (nr, nz)    
    
    
    """
    
    # Read header
    # xmin(cm), xmax(cm), nx-1
    # freq(MHz)
    # ymin(cm), ymax(cm), ny-1
    with open(t7file, 'r') as f:
        line1 = f.readline().split()
        freq_MHz = float(f.readline())
        line3 = f.readline().split()
    
    # Form output dict 
    d = {}
    d['geometry'] = geometry
    d['problem'] = 'fish'        
    d['zmin'], d['zmax'], d['nz'] =  float(line1[0]), float(line1[1]), int(line1[2])+1
    d['freq'] = freq_MHz
    d['rmin'], d['rmax'], d['nr'] =  float(line3[0]), float(line3[1]), int(line3[2])+1
    
    # These should be the labels
    labels=['Ez', 'Er', 'E', 'Hphi']
    
    # Read and reshape
    dat4 = np.loadtxt(t7file, skiprows=3)
    ncol = len(labels)
    dat4 = dat4.reshape(d['nr'], d['nz'], ncol)
    
    
    for i, label in enumerate(labels):
        d[label] = dat4[:,:,i]
    
    
    return d

def parse_poisson_t7(t7file, type='electric', geometry='cylindrical'):
    """
    Parses a T7 file. The T7 header should have:
    
    xmin(cm), xmax(cm), nx-1
    ymin(cm), ymax(cm), ny-1
    
    For type=='electric':
        2 columns of data: Er, Ez
        Units are in V/cm
    
    For type=='magnetic':
        2 columns of data: Br, Bz
        Units are G
    
    Returns a dict with:
        rmin
        rmax
        nr
        ymin
        ymax
        ny
        data: 2D array of shape (nx, ny)    
    
    
    """
    assert geometry == 'cylindrical', 'TODO: other geometries'

    if type == 'electric':
        labels = 'Er', 'Ez'
    elif type == 'magnetic':
        labels = 'Br', 'Bz'
        
    # Read header
    # xmin(cm), xmax(cm), nx-1    # r in cylindrical geometry
    # ymin(cm), ymax(cm), ny-1    # z in cylindrical geometry
    with open(t7file, 'r') as f:
        xline = f.readline().split()
        yline = f.readline().split()

    # Form output dict 
    d = {}
    d['geometry'] = geometry
    d['problem'] = 'poisson'        
    d['rmin'], d['rmax'], d['nr'] =  float(xline[0]), float(xline[1]), int(xline[2])+1
    d['zmin'], d['zmax'], d['nz'] =  float(yline[0]), float(yline[1]), int(yline[2])+1
    
    # Read and reshape
    dat4 = np.loadtxt(t7file, skiprows=2)
    ncol = len(labels)
    dat4 = dat4.reshape(d['nz'], d['nr'], ncol)
    
    for i, label in enumerate(labels):
        d[label] = dat4[:,:,i].T
    
    
    return d



#_________________________________
#_________________________________
# Individual parsers



#_________________________________
# Header

def parse_header_variable(line):
    """
    Parses a line that follows:
    
    Variable Code         Value     Description
    
    Returns:
        key, value, description, in_automesh
    
    """
    x = line.split()
    
    key = x[0]
    
    if x[1] == 'A':
        in_automesh = True
        
        s = x[2]
        d = x[3:]
    else:
        in_automesh = False
        s = x[1]
        d = x[2:]
        
    descrip = ' '.join(d)
        
    try: 
        val = int(s)
    except ValueError:
        val = float(s)
    
    return key, val, descrip, in_automesh


def parse_header_lines(lines):
    """
    Parses the header lines
    """
    
    
    header = 'Variable Code         Value     Description'
    
    d = {}
    description = {}
    in_automesh = {}
    
    comments = []
    
    in_header=False
    for line in lines:
        if line == header:
            in_header = True
            continue
        if not in_header:
            comments.append(line)
            continue
        
        key, val, descrip, in_am = parse_header_variable(line)
        
        d[key] = val
        description[key] = descrip
        in_automesh[key] = in_am
        
    return {'variable':d, 'description':description, 'in_automesh':in_automesh, 'comments':'\n'.join(comments)}


#_________________________________
# Wall segments




def parse_wall_segment_line1(line):
    """
    helper parse_sfo_segment
    """
    d = {}
    ix, x = line.split('segment')[1].split(' K,L =')
    d['segment_number'] = int(float((ix)))
    kl0, kl1 = x.split('to')
    k0, l0 = kl0.split(',')
    d['K_beg'], d['L_beg'] = int(k0), int(l0)
    k1, l1 = kl1.split(',')
    d['K_end'], d['L_end'] = int(k1), int(l1)
    return d


def parse_sfo_segment(lines):
    """
    Parses lines that start with:
        'Power and fields on wall segment'
    """

    # key = value lines

    info = parse_wall_segment_line1(lines[0])
    
    inside = False
    fields = None
    units = None

    for L in lines[1:]:

        L = L.strip()

        # Look for key=value
        if '=' in L:
            key, val = L.split('=')
            info[key.strip()] = val
            continue

        if L.startswith("K    L"):                
            nskip = 2
            fields = {name.strip('|'):[] for name in L.split()[nskip:]}
            continue
        elif L.startswith( "m     K     L"):
            nskip = 3
            fields = {name.strip('|'):[] for name in L.split()[nskip:]}            
            continue

        if not fields:
            continue        
    
        # Look for units
        if fields and not units:
            unit_labels = L.split()
            assert len(unit_labels) == len(fields), print(unit_labels)
            # make dict
            units = dict(zip(list(fields),unit_labels))
            inside = True
            continue

        # This might come at the end
        if L.startswith('Summary'):
            inside = False
            
            
        # Must be inside. Add data            
        if inside:
            x = [float(y) for y in L.split()]
            # Special care if there are blanks for the skip columns 
            if len(x) == len(fields) + nskip:
                x =  x[nskip:]
            for i, name in enumerate(fields):
                fields[name].append(x[i])

    # Exiting
    for k, v in fields.items():
        fields[k] = np.array(v)            
    
    return {'wall':fields, 'info':info, 'units':units}




#_________________________________
# Summary

def parse_sfo_beam_energy(lines):
    d_vals = {}
    d_units = {}
    for line in lines:
        line = line.strip()
        if line.startswith('V0'):
            line = line.split('=')[-1]
            line = line.strip()
            line = line.split(' ')
            data = line[0]
            data = float(data)
            unit = line[1]
    d_vals['BeamEnergy'] = data
    d_units['BeamEnergy'] = unit
    return d_vals, d_units


def parse_sfo_summary_group(lines):
    """
    """
    d_vals = {}
    d_units = {}
    for line in lines:
        if line == "":
            break
        else:
            d_val, d_unit = parse_sfo_summary_group_line(line)
            d_vals.update(d_val)
            d_units.update(d_unit)
    return d_vals, d_units

def parse_simple_summary_line(line):
    # deal with simple line with one key and one value
    d_val = {}
    d_unit = {}
    line = line.split("=")
    if len(line) == 1:
        return d_val, d_unit
    key = line[0].strip()
    val = line[-1]
    val = val.strip()
    val = val.split(" ")
    d_val[key] = float(val[0])
    if len(val) > 1:
        d_unit[key] = val[1]
    else:
        d_unit[key] = ""
    return d_val, d_unit

def parse_sfo_summary_group_line(line):
    d_val = {}
    d_unit = {}
    if line.startswith('Field normalization'):
        line = line.split("=")
        val = line[-1]
        val = val.strip()
        val = val.split(" ")
        d_val['Enorm'] = float(val[0])
        d_unit['Enorm'] = val[1]
        return d_val, d_unit
# 'for the integration path from point Z1,R1 =     50.50000 cm,   0.00000 cm',
    if line.startswith('for the integration path'):
        line = line.split("=")
        val = line[-1]
        val = val.strip()
        val = val.split(",")
        v1 = val[0].split(" ")
        v1 = v1[0]
        v2 = val[1].strip()
        v2 = v2.split(" ")
        v2 = v2[0]
        d_val['integration_Z1'] = float(v1)
        d_unit['integration_Z1'] = "cm"
        d_val['integration_R1'] = float(v2)
        d_unit['integration_R1'] = "cm"
        return d_val, d_unit
#  'to ending point                     Z2,R2 =     50.51000 cm,   0.00000 cm',
    if line.startswith('to ending point'):
        line = line.split("=")
        val = line[-1]
        val = val.strip()
        val = val.split(",")
        v1 = val[0].split(" ")
        v1 = v1[0]
        v2 = val[1].strip()
        v2 = v2.split(" ")
        v2 = v2[0]
        d_val['integration_Z2'] = float(v1)
        d_unit['integration_Z2'] = "cm"
        d_val['integration_R2'] = float(v2)
        d_unit['integration_R2'] = "cm"
        return d_val, d_unit
    if line.startswith('Beta '):
        # custom parse the beta line
        line = line.split("=")
        beta = line[1].strip()
        beta = beta.split(" ")
        beta = beta[0]
        ke = line[-1]
        ke = ke.strip()
        ke = ke.split(" ")
        ke = ke[0]
        d_val['beta'] = float(beta)
        d_val['kinetic_energy'] = float(ke)
        d_unit['beta'] = ""
        d_unit['kinetic_energy'] = "MeV"
        return d_val, d_unit
#   'Q    =  0.334933E+10      Shunt impedance =  2001715.397 MOhm/m',
    if line.startswith('Q '):
        # Q factor line
        line = line.split("=")
        v1 = line[1].strip()
        v1 = v1.split(" ")
        v1 = v1[0]
        v2 = line[-1]
        v2 = v2.strip()
        v2 = v2.split(" ")
        v2 = v2[0]
        d_val['Q'] = float(v1)
        d_val['Shunt impedance'] = float(v2)
        d_unit['Q'] = ""
        d_unit['Shunt impedance'] = "MOhm/m"
        return d_val, d_unit
#          'Rs*Q =    81.364 Ohm                Z*T*T =  1956056.397 MOhm/m',
    if line.startswith('Rs*Q '):
        # Q factor line
        line = line.split("=")
        v1 = line[1].strip()
        v1 = v1.split(" ")
        v1 = v1[0]
        v2 = line[-1]
        v2 = v2.strip()
        v2 = v2.split(" ")
        v2 = v2[0]
        d_val['Rs*Q'] = float(v1)
        d_val['Z*T*T'] = float(v2)
        d_unit['Rs*Q'] = "Ohm"
        d_unit['Z*T*T'] = "MOhm/m"
        return d_val, d_unit
# 'r/Q  =   134.323 Ohm  Wake loss parameter =      0.04044 V/pC',
    if line.startswith('r/Q '):
        # r/Q line
        line = line.split("=")
        v1 = line[1].strip()
        v1 = v1.split(" ")
        v1 = v1[0]
        v2 = line[-1]
        v2 = v2.strip()
        v2 = v2.split(" ")
        v2 = v2[0]
        d_val['r/Q'] = float(v1)
        d_val['Wake loss parameter'] = float(v2)
        d_unit['r/Q'] = "Ohm"
        d_unit['Wake loss parameter'] = "V/pC"
        return d_val, d_unit
#  'Average magnetic field on the outer wall  =      16678.8 A/m, 0.337887 mW/cm^2',
    if line.startswith('Average magnetic '):
        # Average magnetic field line
        line = line.split("=")
        val = line[-1]
        val = val.strip()
        val = val.split(",")
        v1 = val[0].split(" ")
        v1 = v1[0]
        d_val['AvgH'] = float(v1)
        d_unit['AvgH'] = "A/m"
        return d_val, d_unit
#  'Maximum H (at Z,R = 25.1487,18.4727)      =       41189. A/m, 2.06066 mW/cm^2',
    if line.startswith('Maximum H '):
        # Maximum H line
        line = line.split("=")
        val = line[1].strip()
        val = val.split(",")
        v1 = val[0]
        v2 = val[1].split(")")
        v2 = v2[0]
        v3 = line[-1]
        v3 = v3.strip()
        v3 = v3.split(",")
        v3 = v3[0].split(" ")
        v3 = v3[0]
        d_val['MaxH_z'] = float(v1)
        d_val['MaxH_r'] = float(v2)
        d_val['MaxH'] = float(v3)
        d_unit['MaxH_z'] = 'cm'
        d_unit['MaxH_r'] = 'cm'
        d_unit['MaxH'] = "A/m"
        return d_val, d_unit
#  'Maximum E (at Z,R = 49.9969,0.624269)     =      41.1564 MV/m, 2.84161 Kilp.',
    if line.startswith('Maximum E '):
        # Maximum E line
        line = line.split("=")
        val = line[1].strip()
        val = val.split(",")
        v1 = val[0]
        v2 = val[1].split(")")
        v2 = v2[0]
        v3 = line[-1]
        v3 = v3.strip()
        v3 = v3.split(",")
        v3 = v3[0].split(" ")
        v3 = v3[0]
        d_val['MaxE_z'] = float(v1)
        d_val['MaxE_r'] = float(v2)
        d_val['MaxE'] = float(v3)
        d_unit['MaxE_z'] = 'cm'
        d_unit['MaxE_r'] = 'cm'
        d_unit['MaxE'] = "MV/m"
        return d_val, d_unit
    else:
        d_val, d_unit = parse_simple_summary_line(line)
    return d_val, d_unit



from setuptools import setup, find_packages
from os import path, environ

cur_dir = path.abspath(path.dirname(__file__))

with open(path.join(cur_dir, 'requirements.txt'), 'r') as f:
    requirements = f.read().split()



setup(
    name='pySuperfish',
    version = '0.0.1',
    packages=find_packages(),  
    package_dir={'superfish':'superfish'},
    url='https://github.com/ChristopherMayes/pySuperfish',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    install_requires=requirements,
    include_package_data=True,
    python_requires='>=3.6'
)
